@@ -4,23 +4,19 @@
 # LICENSE file in the root directory.
 
 """ Numpy based implementation of multi objective
-    Differential Evolution using the DE/pareto/1 strategy. 
-    Derived and adapted for MO from its C++ counterpart 
-    https://github.com/dietmarwo/fast-cma-es/blob/master/_fcmaescpp/deoptimizer.cpp
-    
+    Differential Evolution using the DE/all/1 strategy. 
+     
     Can switch to NSGA-II like population update via parameter 'nsga_update'.
     Then it works essentially like NSGA-II but instead of the tournament selection
     the whole population is sorted and the best individuals survive. To do this
     efficiently the crowd distance ordering is slightly inaccurate - each objective
     is ordered separately. 
     
-    Compared to most Python NSGA-II implementations it supports parallel fitness function 
-    evaluation. 
+    Supports parallel fitness function evaluation. 
     
     Enables the comparison of DE and NSGA-II population update mechanism with everything else
     kept completely identical.
     
-    
     Requires python 3.5 or higher. 
     
     Uses the following deviation from the standard DE algorithm:
@@ -30,11 +26,12 @@
     since the algorithm oscillates between different F and CR settings. 
     
     For expensive objective functions (e.g. machine learning parameter optimization) use the workers
-    parameter to parallelize objective function evaluation. This causes delayed population update.
-    It is usually preferrable if popsize > workers and workers = mp.cpu_count() to improve CPU utilization.  
+    parameter to parallelize objective function evaluation. The workers parameter is limited by the 
+    population size.
 """
 
 import numpy as np
+import os
 import sys
 import time
 import ctypes as ct
@@ -45,18 +42,23 @@ import multiprocessing as mp
 from fcmaes.optimizer import dtime, logger
 from fcmaes import retry, moretry
 
+os.environ['MKL_DEBUG_CPU_TYPE'] = '5'
+
 def minimize(mofun, 
              nobj,
              dim = None,
-             bounds = None, 
+             bounds = None,
+             ncon = 0, 
              popsize = 64, 
              max_evaluations = 100000, 
              workers = None,
              f = 0.5, 
              cr = 0.9, 
              nsga_update = False,
+             pareto_update = False,
              rg = Generator(MT19937()),
-             logger = None):  
+             logger = None,
+             plot_name = None):  
       
     """Minimization of a multi objjective function of one or more variables using
     Differential Evolution.
@@ -79,6 +81,9 @@ def minimize(mofun,
             1. Instance of the `scipy.Bounds` class.
             2. Sequence of ``(min, max)`` pairs for each element in `x`. None
                is used to specify no bound.
+    ncon : int, optional
+        number of constraints, default is 0. 
+        The objective function needs to return vectors of size nobj + ncon
     popsize : int, optional
         Population size.
     max_evaluations : int, optional
@@ -93,13 +98,17 @@ def minimize(mofun,
         The recombination constant. Should be in the range [0, 1]. 
         In the literature this is also known as the crossover probability.     
     nsga_update = boolean, optional
-        Use of NSGA-II or DE population update.     
+        Use of NSGA-II or DE population update. Default is False    
+    pareto_update = boolean, optional
+        Use the pareto front for population update if nsga_update = False. Default is False     
     rg = numpy.random.Generator, optional
         Random generator for creating random guesses.
     logger : logger, optional
         logger for log output for tell_one, If None, logging
         is switched off. Default is a logger which logs both to stdout and
         appends to a file ``optimizer.log``.
+    trace_plots : trace_plots, optional
+        flag indicating if the pareto front is plotted during the optimization to monitor progress
             
     Returns
     -------
@@ -112,8 +121,8 @@ def minimize(mofun,
         ``success`` a Boolean flag indicating if the optimizer exited successfully. """
 
     
-    de = DE(dim, nobj, bounds, popsize, workers if not workers is None else 0, 
-            f, cr, nsga_update, rg, logger)
+    de = DE(dim, nobj, bounds, ncon, popsize, workers if not workers is None else 0, 
+            f, cr, nsga_update, pareto_update, rg, logger, plot_name)
     try:
         if workers and workers > 1:
             x, y, evals, iterations, stop = de.do_optimize_delayed_update(mofun, max_evaluations, workers)
@@ -126,9 +135,11 @@ def minimize(mofun,
 
 class DE(object):
     
-    def __init__(self, dim, nobj, bounds, popsize = 64, workers = 0,
-                 F = 0.5, Cr = 0.9, nsga_update = False, rg = Generator(MT19937()), logger = None):
+    def __init__(self, dim, nobj, bounds, ncon=0, popsize = 64, workers = 0,
+                 F = 0.5, Cr = 0.9, nsga_update = False, pareto_update = False, 
+                 rg = Generator(MT19937()), logger = None, plot_name = None):
         self.nobj = nobj
+        self.ncon = ncon
         self.dim, self.lower, self.upper = _check_bounds(bounds, dim)
         if popsize is None:
             popsize = 31
@@ -138,17 +149,22 @@ class DE(object):
         self.F0 = F
         self.Cr0 = Cr
         self.nsga_update = nsga_update
+        self.pareto_update = pareto_update
         self.stop = 0
         self.iterations = 0
         self.evals = 0
         self.mutex = mp.Lock()
         self.p = 0
         self.best_p = None
+        self.trace_plots = not plot_name is None
         self._init()
         if not logger is None:
             self.logger = logger
             self.n_evals = mp.RawValue(ct.c_long, 0)
             self.time_0 = time.perf_counter()
+            self.name = plot_name if not plot_name is None else ''
+            self.name += '_mode_' + str(popsize) + '_' + \
+                ('nsga_update' if nsga_update else ('de_best_update' if pareto_update else 'de_all_update'))
         
     def ask(self):
         """ask for one new argument vector.
@@ -203,26 +219,25 @@ class DE(object):
                 c = self.n_evals.value
                 message = '"c/t={0:.2f} c={1:d} t={2:.2f} y={3!s} x={4!s}'.format(
                     c/t, c, t, str(list(y)), str(list(x)))
-                self.logger.info(message)
+                self.logger.info(message)   
+                if self.trace_plots:             
+                    name = self.name + ' '  + str(self.n_evals.value)
+                    np.savez_compressed(name, xs=self.x, ys=self.y) 
+                    moretry.plot(name, self.ncon, self.x, self.y)
 
         return self.stop 
 
-    def big_value(self):
-        inf = np.empty(self.nobj)
-        inf.fill(1E99)
-        return inf
-
     def _init(self):
-        self.x = np.empty([2*self.popsize, self.dim])
-        self.y = np.empty([2*self.popsize, self.nobj])
+        self.x = np.empty((2*self.popsize, self.dim))
+        self.y = np.empty((2*self.popsize, self.nobj + self.ncon))
         for i in range(self.popsize):
             self.x[i] = self._sample()
-            self.y[i] = self.big_value()
+            self.y[i] = np.array([1E99]*(self.nobj + self.ncon))
         
         next_size = 2*(max(self.workers, self.popsize))
         self.done = np.zeros(next_size, dtype=bool)
-        self.nx = np.empty([next_size, self.dim])
-        self.ny = np.empty([next_size, self.nobj])
+        self.nx = np.empty((next_size, self.dim))
+        self.ny = np.empty((next_size, self.nobj + self.ncon))
         self.vx = self.x.copy()
         self.vp = 0
                                     
@@ -238,11 +253,7 @@ class DE(object):
                 self.x[self.popsize + p] = x
                 self.evals += 1
             self.pop_update()
-            if self.evals >= self.max_evals:
-                break
         x, y = filter(self.x, self.y)
-        x = x[:self.popsize]
-        y = y[:self.popsize]
         return x, y, self.evals, self.iterations, self.stop
 
     def do_optimize_delayed_update(self, fun, max_evals, workers=mp.cpu_count()):
@@ -275,12 +286,10 @@ class DE(object):
             
         evaluator.stop()
         x, y = filter(self.x, self.y)
-        x = x[:self.popsize]
-        y = y[:self.popsize]
         return x, y, self.evals, self.iterations, self.stop
 
     def pop_update(self):
-        domination = pareto(self.y)
+        domination = pareto(self.y, self.nobj, self.ncon)
         x = []
         y = []
         maxdom = int(max(domination))
@@ -294,9 +303,11 @@ class DE(object):
                 y = [*y, *self.y[domlevel]]
             else: # sort for crowding
                 nx = self.x[domlevel]
-                ny = self.y[domlevel]                
-                cd = crowd_dist(ny)
-                si = np.flip(np.argsort(cd))
+                ny = self.y[domlevel]    
+                si = [0]
+                if len(ny) > 1:                
+                    cd = crowd_dist(ny)
+                    si = np.flip(np.argsort(cd))
                 for p in si:
                     if len(x) >= self.popsize:
                         break
@@ -319,15 +330,17 @@ class DE(object):
             self.Cr = 0.5*self.Cr0 if self.iterations % 2 == 0 else self.Cr0
             self.F = 0.5*self.F0 if self.iterations % 2 == 0 else self.F0
         while True:
-            if not self.best_p is None and len(self.best_p) > 3: # sample from pareto front
+            if self.pareto_update and (not self.best_p is None) and len(self.best_p) > 3: 
+                # sample from pareto front
                 rb = self.rg.integers(0, len(self.best_p))
                 rb = self.best_p[rb]
                 r1, r2 = self.rg.integers(0, self.popsize, 2)
             else:
+                # sample from whole population
                 r1, r2, rb = self.rg.integers(0, self.popsize, 3)
-            if r1 != p and r1 != rb and r1 != r2 and r2 != rb \
-                    and r2 != p and rb != p:
-                break
+                if r1 != p and r1 != rb and r1 != r2 and r2 != rb \
+                        and r2 != p and rb != p:
+                    break
         xp = self.x[p]
         xb = self.x[rb]
         x1 = self.x[r1]
@@ -367,11 +380,45 @@ def filter(x, y):
     x = np.array(x[:len(y)])
     return x,y
 
-def pareto(ys):
-    dim = len(ys)
-    pareto = np.arange(ys.shape[0])
+def maxcon(cons):
+    cons = np.minimum(cons, 1E99)
+    maxc = max(cons)
+    if maxc == 0:
+        return 0
+    else: # reduce number and value of constraint violations
+        n = sum([c > 0 for c in cons])
+        return n*maxc + sum(cons)
+
+def pareto(ys, nobj, ncon):
+    if ncon == 0:
+        return pareto_levels(ys)
+    else:
+        yobj = np.array([y[:nobj] for y in ys])
+        ycon = np.array([np.maximum(y[-ncon:], 0) for y in ys])    
+        csum = [maxcon(cons) for cons in ycon]
+        ci = np.argsort(csum)
+        popn = len(ys)
+        domination = np.zeros(popn)
+        # first pareto front of feasible solutions
+        cy = np.array([i for i in ci if csum[i] == 0])
+        if len(cy) > 0:
+            ypar = pareto_levels(yobj[cy])
+            domination[cy] += ypar
+        # then constraint violations   
+        ci = np.array([i for i in ci if csum[i] > 0 and csum[i] < 1E99])  
+        if len(ci) > 0:    
+            maxcdom = len(ci)
+            cdom = np.arange(maxcdom, 0, -1)
+            domination[ci] += cdom
+            if len(cy) > 0: # priorize feasible solutions
+                domination[cy] += maxcdom + 1
+    return domination   
+ 
+def pareto_levels(ys):
+    popn = len(ys)
+    pareto = np.arange(popn)
     index = 0  # Next index to search for
-    domination = np.zeros(dim)
+    domination = np.zeros(popn)
     while index < len(ys):
         mask = np.any(ys < ys[index], axis=1)
         mask[index] = True
@@ -381,25 +428,22 @@ def pareto(ys):
         index = np.sum(mask[:index])+1
     return domination
 
-def crowd_dist(y): # scaled crowd distance
-    objs = zip(*y)
-    dist = np.zeros(len(y))
-    for obj in objs:
-        obj = np.array(obj)
-        if max(obj) > 1E10: # no scaling possible
-            continue;
-        si = np.argsort(obj)
-        scale = obj[si[-1]] - obj[si[0]] # scale
-        if scale <= 0:
-            return np.arange(len(y))
-        prev = -1
-        for i in si:
-            if prev >= 0:
-                d = abs(obj[i] - obj[prev]) / scale 
-                dist[prev] += d
-                dist[i] += d
-            prev = i 
-    return dist
+def crowd_dist(y): # crowd distance for 1st objective
+    n = len(y)
+    y0 = np.array([yi[0] for yi in y])
+    si = np.argsort(y0) # sort 1st objective
+    y0_s = y0[si] # sorted
+    d = y0_s[1:n] - y0_s[0:n-1] # neighbor distance
+    if max(d) == 0:
+        return np.zeros(n)
+    dsum = np.zeros(n)
+    dsum += np.array(list(d) + [0]) # distance to left
+    dsum += np.array([0] + list(d)) # distance to right
+    dsum[0] = 1E99 # keep borders
+    dsum[-1] = 1E99
+    ds = np.empty(n)
+    ds[si] = dsum # inverse order
+    return ds
 
 # derived from https://github.com/ChengHust/NSGA-II/blob/master/GLOBAL.py
 def variation(pop, lower, upper, rg, pro_c = 1, dis_c = 20, pro_m = 1, dis_m = 20):
@@ -414,7 +458,8 @@ def variation(pop, lower, upper, rg, pro_c = 1, dis_c = 20, pro_m = 1, dis_m = 2
     beta[mu > 0.5] = np.power(2 * mu[mu > 0.5], -1 / (dis_c + 1))
     beta = beta * ((-1)** rg.integers(2, size=(n // 2, d)))
     beta[rg.random((n // 2, d)) < 0.5] = 1
-    beta[np.tile(rg.random((n // 2, 1)) > pro_c, (1, d))] = 1
+    if pro_c < 1.0:
+        beta[np.tile(rg.random((n // 2, 1)) > pro_c, (1, d))] = 1
     offspring = np.vstack(((parent_1 + parent_2) / 2 + beta * (parent_1 - parent_2) / 2,
                                (parent_1 + parent_2) / 2 - beta * (parent_1 - parent_2) / 2))
     site = rg.random((n, d)) < pro_m / d
@@ -434,20 +479,15 @@ def variation(pop, lower, upper, rg, pro_c = 1, dis_c = 20, pro_m = 1, dis_m = 2
     offspring = np.maximum(np.minimum(offspring, upper), lower)
     return offspring
 
-def minimize_plot(name, fun, bounds, nobj, popsize = 64, max_eval = 100000, nsga_update=False, 
-                  workers = mp.cpu_count(), logger=logger()):
-    time0 = time.perf_counter() # optimization start time
-    name += '_mode_' + str(popsize) + '_' + str(max_eval) + '_' + ('nsga_update' if nsga_update else 'de_update')
+def minimize_plot(name, fun, bounds, nobj, ncon = 0, popsize = 64, max_eval = 100000, nsga_update=False, 
+                  pareto_update=False, workers = mp.cpu_count(), logger=logger(), plot_name = None):
+    name += '_mode_' + str(popsize) + '_' + \
+                ('nsga_update' if nsga_update else ('de_best_update' if pareto_update else 'de_all_update'))
     logger.info('optimize ' + name) 
-    ret = minimize(fun, nobj, bounds = bounds, popsize = popsize, max_evaluations = max_eval,
-                   nsga_update = nsga_update, workers=workers, logger=logger)
+    ret = minimize(fun, nobj, bounds = bounds, ncon = ncon, popsize = popsize, max_evaluations = max_eval,
+                   nsga_update = nsga_update, pareto_update = pareto_update, workers=workers, 
+                   logger=logger, plot_name = plot_name)
     ys = np.array(ret.fun)
     xs = np.array(ret.x)
-    retry.plot(ys, 'all_' + name + '.png', interp=False)
     np.savez_compressed(name, xs=xs, ys=ys)
-    xs, front = moretry.pareto(xs, ys)
-    logger.info(name + ' time ' + str(dtime(time0))) 
-    retry.plot(front, 'front_.' + name + '.png')
-
-
-        
+    moretry.plot(name, ncon, xs, ys)
